import nopickles
import renpy
import sanitize

# This section might be refactored back into sanitize.py, it's not really
# bc-specific.
def alnum_lc(s):
    ret = ''
    for c in s:
        if c not in 'abcdefghijklmnopqrstuvwxyz0123456789':
            raise Exception("Only lowercase letters and numbers are permitted.")
        ret += c
    return ret

def alpha_lc(s):
    ret = ''
    for c in s:
        if c not in 'abcdefghijklmnopqrstuvwxyz':
            raise Exception("Only lowercase letters are permitted.")
        ret += c
    return ret

def alpha_lc_or_None(s):
    if s is None:
        return None
    elif (type(s) is not str) and (type(s) is not unicode):
        nopickles.warnings.append((5,'expected lowercase letters; discarded',s))
        return None
    return alpha_lc(s)

def expect_int_or_None(i):
    if i is None:
        return i
    return int(i)

def expect_int_or_False(i):
    if type(i) is int:
        return i
    if i is False:
        return False
    raise Exception("Expected int or False.")

def expect_int3(s):
    if (type(s) is not tuple) or (len(s)!=3) or (type(s[0]) is not int) or (type(s[1]) is not int) or (type(s[2]) is not int):
        nopickles.warnings.append((5,'Expected an int3, discarded',type(s),s))
        return None
    return s

def expect_MissTalmage(s):
    if type(s) is renpy.store.MissTalmage:
        return s
    raise Exception('Expected Miss Talmage.')

def expect_RecordsOffice(s):
    if type(s) is renpy.store.RecordsOffice:
        return s
    raise Exception('Expected a RecordsOffice.')

def expect_RevertableDict(s):
    if not nopickles.INSECURE:
        raise Exception("BAD!")
    if isinstance(s,renpy.python.RevertableDict):
        nopickles.warnings.append((6,"Passing RevertableDict",s))
        return s
    raise Exception("Expected a RevertableDict.")

def expect_rlist_of_instances(l, t):
    """
    Return a new RevertableList containing all instances of t (or subclasses
    of t) in RevertableList l.
    """
    if type(l) is not renpy.python.RevertableList:
        nopickles.warnings.append((5,"Expected a RevertableList, got",type(l),l))
        return renpy.python.RevertableList()
    ret = renpy.python.RevertableList()
    for i in l:
        if isinstance(i,t):
            ret.append(i)
        else:
            nopickles.warnings.append((5, "Expected an instance of",t,'in RevertableList(), got',type(i),l))
    return ret

def expect_RevertableList_of_str_or_unicode(s):
    if not nopickles.INSECURE:
        raise Exception("Unaudited section.")
    ret = renpy.python.RevertableList()
    if not isinstance(s, renpy.python.RevertableList):
        nopickles.warnings.append((5,"Expected a RevertableList, got",type(s),s))
        return ret
    for item in s:
        if (type(item) is str) or (type(item) is unicode):
            ret.append(item)
        else:
            nopickles.warnings.append((5,'Expected a RevertableList of str or unicode, discarded',type(item),item))
    return ret

def expect_str_numeric(s):
    ret = ''
    for c in s:
        if c not in '0123456789':
            raise Exception("Expected a digit.")
        ret += c
    return ret

def expects_instance(t):
    """
    Returns a type-checking function; if the type check fails, the type-checker
    will return None.
    """
    def ret(s):
        if isinstance(s,t):
            return s
        nopickles.warnings.append((5,s,"is not a",t,"so it was discarded."))
        return None
    return ret

def expects_tuple_of(t, count):
    def ret(s):
        if (type(s) is not tuple) or (len(s)!=count):
            nopickles.warnings.append((5, "Expected tuple of",t,count,'got',s))
            return (0,)*count
        for item in s:
            if type(item) is not t:
                nopickles.warnings.append((5,"Expected tuple of",t,'but got an',type(item)))
                return (0,)*count
        return s
    return ret

# Begin really bc-specific stuff.
# Used to clear out seen images on persistent serialization.
DISCARD_SEEN_IMAGES = True

# alphabetized list of sanitizers

def expect_Actor_instance_or_None(s):
    if isinstance(s, renpy.store.Actor):
        return s
    if s is not None:
        nopickles.warnings.append((5,'expect_Actor_instance_or_None discarded',s))
    return None

def expect_Bedroom_instance(loc):
    if not isinstance(loc, renpy.store.Bedroom):
        raise Exception("Expected a Bedroom instance.")
    return loc

def expect_Bedroom_or_None(loc):
    if loc is None:
        return None
    if isinstance(loc,renpy.store.Bedroom):
        return loc
    raise Exception("A bedroom or None is required.")

def expect_Case_instance_or_del(c):
    if isinstance(c,renpy.store.Case):
        return c
    if type(c) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5,'expected case or del, got',c,type(c)))
    return renpy.python.deleted

def expect_Case_instance_or_del_or_None(c):
    if c is None:
        return None
    return expect_Case_instance_or_del(c)

def expect_Case_instance_or_None(c):
    if c is None:
        return None
    if isinstance(c,renpy.store.Case):
        return c
    nopickles.warnings.append((5,'expect_Case_instance_or_None',type(c),c))
    return None

def expect_Caseref(c):
    if type(c) is renpy.store.Caseref:
        return c
    nopickles.warnings.append((5,'expected caseref, discarded',c))
    return None

def expect_Caseref_or_None(c):
    if type(c) is renpy.store.Caseref:
        return c
    if c is not None:
        nopickles.warnings.append((5,'expected caseref, discarded',c))
    return None

def expect_Case_instance(s):
    return sanitize.expect_instance(s, renpy.store.Case)

def expect_Councillor(s):
    if isinstance(s, renpy.store.Councillor):
        return s
    nopickles.warnings.append((5,'discarded non-councillor value',type(s)))
    return None

def expect_Councillor_or_del(s):
    if isinstance(s, renpy.store.Councillor):
        return s
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5,'discarded non-councillor non-del value',type(s)))
    return renpy.python.deleted

def expect_Councillor_or_None(s):
    if isinstance(s,renpy.store.Councillor):
        return s
    if s is not None:
        nopickles.warnings.append((5,'expect_Councillor_or_None discarded',s))
    return None

def expect_councillor_name(s):
    if s in ('Althea','Vonne','Mallory','Rowan'):
        return str(s)
    if s == u'Thaïs':
        return u'Thaïs'
    nopickles.warnings.append((5,"Discarded unexpected councillor name",s))
    return None

def expect_councillor_name_or_del(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    return expect_councillor_name(s)

def expect_del_or_Location_instance(s):
    if isinstance(s, renpy.store.Location):
        return s
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5,'Expected a Location or del, discarded',type(s)))
    return renpy.python.deleted

def expect_del_or_rlist_of_Case_instances(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    return expect_rlist_of_instances(s, renpy.store.Case)

def expect_del_or_Student_instance(s):
    if isinstance(s, renpy.store.Student):
        return s
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5, 'expect_del_or_Student_instance discarded',s,type(s)))
    return renpy.python.deleted

def expect_del_or_rlist_of_Location_instances(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    return expect_rlist_of_instances(s,renpy.store.Location)

def expect_Location_instance(s):
    if isinstance(s, renpy.store.Location):
        return s
    nopickles.warnings.append((5,'Expected a Location, discarded',type(s)))
    return None

def expect_Location_instance_or_None(s):
    if isinstance(s, renpy.store.Location):
        return s
    elif s is not None:
        nopickles.warnings.append((5,'Expected a Location or None, discarded',type(s)))
    return None

def expect_rlist_of_Actor_instances(s):
    ret = renpy.python.RevertableList()
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    if type(s) is not renpy.python.RevertableList:
        nopickles.warnings.append((5,'discarded rlist_of_Actor',s))
        return ret
    for item in s:
        item = nopickles.deindirect_if_needed(item)
        if isinstance(item,renpy.store.Actor):
            ret.append(item)
        else:
            nopickles.warnings.append((5,'expected an Actor instance, discarded',item))
    return ret

def expect_rlist_of_Case_instances(s):
    return expect_rlist_of_instances(s, renpy.store.Case)

def expect_rlist_of_Case_instances_or_del_or_etuple(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    if type(s) is tuple and len(s)==0:
        return ()
    return expect_rlist_of_Case_instances(s)

def expect_rlist_of_Case_instances_or_etuple(s):
    if type(s) is tuple and len(s)==0:
        return ()
    return expect_rlist_of_Case_instances(s)

def expect_rlist_of_Location_instances(s):
    return expect_rlist_of_instances(s,renpy.store.Location)

def expect_rlist_of_Student(l):
    ret = renpy.python.RevertableList()
    if type(l) is renpy.python.RevertableList:
        warned = False
        for i in l:
            if type(i) is renpy.store.Student:
                ret.append(i)
            else:
                warned = True
                nopickles.warnings.append((5, "Expected a RevertableList of Student, discarded", type(i)))
                if type(i) is renpy.store.DummyStudent:
                    nopickles.warnings.append((5,'dummydict',i.__dict__,i._case.serial))
        if warned:
            nopickles.warnings.append((5,"original list was",l))
        return ret
    nopickles.warnings.append((5,"Expected a RevertableList, discarded",type(l)))
    return ret

def expect_rlist_of_Student_instances(l):
    if type(l) is renpy.python.RevertableList:
        ret = renpy.python.RevertableList()
        for i in l:
            if isinstance(i,renpy.store.Student):
                ret.append(i)
            else:
                nopickles.warnings.append((5, "Expected a RevertableList of Student instances, discarded",type(i),'from',l))
                if type(i) is renpy.store.DummyStudent:
                    nopickles.warnings.append((5,'dummydict',i.__dict__,i._case.serial))
        return ret
    nopickles.warnings.append((5,"Expected a RevertableList, discarded",type(l)))
    return renpy.python.RevertableList()

def expect_Student_instance(s):
    if isinstance(s,renpy.store.Student):
        return s
    nopickles.warnings.append((5,'Expected a Student, discarded',type(s)))
    return None

def expect_Student_instance_or_None(s):
    if s is None:
        return s
    if isinstance(s,renpy.store.Student):
        return s
    nopickles.warnings.append((5,"Expected a Student or None, discarded",type(s)))
    return None

def scrub__a_trace_of_scarlet(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    elif type(s) is renpy.store.ATraceOfScarlet:
        return s
    else:
        raise Exception('Invalid value for _a_trace_of_scarlet.')

OK_ACTIONS = set(('Attend','Confront','Detention','Expel','Explore','Get Supplies','Give Presentation','Go Alone','Gossip','Guard','Harass','Lemonade Stand','Mislead','Persuade','Plant','Question','Search','Send Signal','Set A Trap','Stake Out','Stalk','Suspend','Train'))

def scrub_actions(l):
    # Students and Locations must have this sanitizer.
    ret = []
    for a in l:
        if a in OK_ACTIONS:
            ret.append(str(a))
        else:
            nopickles.warnings.append((5,'Discarded action',a))
    return renpy.python.RevertableList(ret)

def scrub_actors(l):
    # Locations and Students should both have this.
    ret = []
    for a in l:
        a = nopickles.deindirect_if_needed(a)
        # Student to accomodate tutorial and future plotless mode.
        if isinstance(a,renpy.store.Student) or isinstance(a,renpy.store.InventoryItem) or isinstance(a,renpy.store.MissTalmage):
            ret.append(a)
        else:
            print "Debug: a",a
            raise Exception("Unexpected value in actors.")
    return renpy.python.RevertableList(ret)

def scrub_all_students_ever(s):
    if type(s) is not renpy.python.RevertableDict:
        if type(s) is renpy.python.StoreDeleted:
            return renpy.python.deleted
        nopickles.warnings.append((5,'Discarded non-RevertableDict all_students_ever value',type(s)))
        return renpy.python.RevertableDict()
    ret = renpy.python.RevertableDict()
    for k,v in s.iteritems():
        ret[(sanitize.expect_displayable_text(k[0]),sanitize.expect_int(k[1]))] = expect_Student_instance(v)
    return ret

def scrub_althea_caught_up_with_fangirl(s):
    if type(s) is str:
        return sanitize.expect_displayable_text(s)
    if type(s) is bool:
        return s
    if s is not None:
        nopickles.warnings.append((5,'Discarded invalid althea_caught_up_with_fangirl',s))
    return None

def scrub_althea_success(s):
    if s in ('romantic','non-romantic'):
        return str(s)
    if type(althea_success) is bool:
        return althea_success
    if s is not None:
        nopickles.warnings.append((5,'Discarded invalid althea_success',s))
    return None

def scrub_Case_history(hist):
    ret = []
    for case in hist:
        if type(case) is renpy.store.Caseref:
            ret.append(case)
        elif isinstance(case, renpy.store.Case):
            ret.append(case)
        else:
            nopickles.warnings.append((5,"Expected Caseref or Case, got",case))
    return renpy.python.RevertableList(ret)

def scrub_Case_log_paginated(lp):
    ret = renpy.python.RevertableList()
    if type(lp) is not renpy.python.RevertableList:
        nopickles.warnings.append((5,'Discarded invalid paginated log',type(lp),lp))
        return ret
    # The paginated log is a list of pages.
    #   Each page is a list of entries. 
    #       Each entry is a list of [dtuple, unicode]
    #           Each dtuple is (y, m, d, wday, [continued])
    nopickles.warnings.note('Case.sanitize_log_paginated')
    for page in lp:
        newpage = []
        for entry in page:
            if len(entry)==2:
                newpage.append(renpy.python.RevertableList(
                    (scrub_Case_log_paginated_dtuple(entry[0]),
                    sanitize.expect_displayable_text(entry[1]))))
            else:
                if nopickles.INSECURE:
                    nopickles.warnings.append((6,"Accepting non-2list",entry,len(entry)))
                    newpage.append(entry)
                else:
                    nopickles.warnings.append((5,"Discarded invalid log entry",entry,len(entry)))
        ret.append(renpy.python.RevertableList(newpage))
    nopickles.warnings.unnote()
    return ret

def scrub_Case_log_paginated_dtuple(dt):
    # replaces the pickle inspector's Case_sanitize_dtuple()
    if len(dt)<4:
        nopickles.warnings.append((5,"Case_sanitize_dtuple discarding invalid",dt))
        return (0,0,0,0)
    elif not (type(dt[0]) is int and type(dt[1]) is int and type(dt[2]) is int and type(dt[3]) is int):
        nopickles.warnings.append((5,"Case_sanitize_dtuple discarding invalid",dt))
        return (0,0,0,0)
    elif len(dt)==5 and type(dt[4] is str):
        return (dt[0],dt[1],dt[2],dt[3],sanitize.expect_displayable_text(dt[4]))
    elif len(dt)>4:
        nopickles.warnings.append((5,"Case_sanitize_dtuple discarding invalid",dt))
        return (0,0,0,0)
    return (dt[0], dt[1], dt[2], dt[3])

def scrub_cases_last_week(s):
    if type(s) is tuple and len(s)==0:
        return ()
    return expect_rlist_of_Case_instances(s)

def scrub_college_acceptances(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    ret = renpy.python.RevertableSet()
    if type(s) is not renpy.python.RevertableSet:
        nopickles.warnings.append((5,'discarded non-set college_acceptances',s))
        return ret
    for k in s:
        ret.add(sanitize.expect_displayable_text(k))
    return ret

def scrub_Councillor_feels(feels):
    if type(feels) is not renpy.python.RevertableDict:
        nopickles.warnings.append((5,"Debug: feels type",type(feels)))
    ret = renpy.python.RevertableDict()
    for k,v in feels.iteritems():
        if type(v) is not int:
            nopickles.warnings.append((5,"Discarded invalid value in feels",v))
            v = 0
        if type(k) is tuple and len(k)==2 and type(k[1])==int and type(k[0]) in (unicode, str):
            # >= 1.0.22, feels should always have a serial number.
            ret[k] = v
        elif (type(k) is str or type(k) is unicode):
            # This is for pre-1.0.22 saves.  Since plotless mode isn't
            # player-visible yet, we can just assume that these feels refer
            # to other plot-mode councillors, who will all have the surname
            # serial 0.
            ret[(k,0)] = v
        else:
            nopickles.warnings.append((5,"Bad type in feels,",k,feels[k]))
    return ret

def scrub_Councillor_on_break(s):
    if type(s) is bool:
        return s
    if type(s) is str or type(s) is unicode:
        return sanitize.expect_displayable_text(s)
    nopickles.warnings.append((5,"boolifying Councillor on_break",s))
    return bool(s)

def scrub_current_action(s):
    if s is None:
        return s
    if s in OK_ACTIONS:
        return str(s)
    nopickles.warnings.append((5,'Discarding unexpected current_action',s))
    return None

def scrub_current_actors(s):
    ret = renpy.python.RevertableList()
    if type(s) is not renpy.python.RevertableList:
        nopickles.warnings.append((5,'discarded invalid current_actors',type(s)))
        return ret
    for actor in s:
        if isinstance(actor,renpy.store.Councillor) or isinstance(actor, renpy.store.InventoryItem):
            ret.append(actor)
    return ret

def scrub_grades(d):
    ret = renpy.python.RevertableDict()
    if type(d) is not renpy.python.RevertableDict:
        if type(d) is renpy.python.StoreDeleted:
            return renpy.python.deleted
        nopickles.warnings.append((5,'Discarded invalid grades',d))
        return ret
    for k,v in d.iteritems():
        if k not in ('A','B','C','D','E','F','*'):
            nopickles.warnings.append((5,'Discarded invalid grade',k,v))
        ret[k] = int(v)
    return ret

def scrub_inventory(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    ret = renpy.python.RevertableList()
    for i in s:
        if isinstance(i,renpy.store.InventoryItem):
            ret.append(i)
        else:
            nopickles.warnings.append((5,'Discarded invalid inventory item',i))
    return ret

def scrub_last_boss_attack(s):
    if s in ('Social','Intimidation','Observation','Stealth','Victory','Surrender'):
        return str(s)
    nopickles.warnings.append((5,'Discarded invalid last_boss_attack',s))
    return 'Victory'

def scrub_locations(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    ret = renpy.python.RevertableDict()
    if type(s) is not renpy.python.RevertableDict:
        nopickles.warnings.append((5,'discarded non-revertabledict locations',s))
        return ret
    for k,v in s.iteritems():
        if isinstance(v, renpy.store.Location):
            ret[sanitize.expect_displayable_text(k)] = v
        else:
            nopickles.warnings.append((5,'Discarded non-Location location',k,v))
    return ret

def scrub_metamorphoses(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    ret = renpy.python.RevertableList()
    for m in s:
        if type(m) is not tuple or len(m)!=4:
            continue
        if type(m[0]) is not float or type(m[1]) is not int or type(m[2]) is not int:
            continue
        # Valid metamorphosis targets all need to be checked here, if other
        # metamorphoses are added.
        if m[3]!='grubslave':
            continue
        ret.append(m)
    return ret

def scrub_minions(s):
    a = []
    if type(s) is not tuple:
        if type(s) is not renpy.python.StoreDeleted:
            nopickles.warnings.append((5,'discarded non-tuple, non-del minions',s))
        return renpy.python.deleted
    for minion in s:
        if isinstance(minion, renpy.store.Student):
            a.append(minion)
        else:
            nopickles.warnings.append((5,'discarded non-Councillor minion',s))
    return tuple(a)

def scrub_mole_case(s):
    if isinstance(s, renpy.store.AMountainOrAMolehill):
        return s
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    if s is None:
        return None
    nopickles.warnings.append((5,'discarded mole_case',s))
    return None

def scrub_mole_selected(s):
    if s == 'None':
        return 'None'
    return expect_councillor_name_or_del(s)

def scrub_overlay_pos(s):
    if isinstance(s,tuple) and len(s)==2:
        return int(s[0]),int(s[1])
    elif s is None:
        return None
    raise Exception("Unexpected overlay pos.")

def scrub_pending_college_acceptances(s):
    if s is False:
        return False
    return sanitize.expect_rlist_of_displayable_text(s)

def scrub_result(s):
    if s=='done':
        return 'done'
    elif type(s) is int:
        return s
    elif s in ('NOTE','ERR','OK','FIXED'):
        return str(s)
    elif type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    nopickles.warnings.append((5,'discarded result',s))
    return None

def scrub_rowan_path(s):
    if type(s) is bool:
        return s
    elif s=='beta':
        return 'beta'
    elif s is not None:
        nopickles.warnings.append((5,'discarded rowan_path',s))
    return None

def scrub_saturday_date(s):
    if isinstance(s, renpy.python.StoreDeleted):
        return renpy.python.deleted
    elif s=='':
        return s
    else:
        return expect_councillor_name(s)

def scrub_scapegoat_(s):
    if s is None:
        return None
    return expect_Councillor_or_del(s)

def scrub_Student_filename(s):
    try:
        return sanitize.scrub_relative_path(s,ok='-.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz')
    except Exception,e:
        nopickles.warnings.append((5,"Discarded student filename:",e))
        return None

def scrub_students_in_play(s):
    ret = renpy.python.RevertableDict()
    if type(s) is not renpy.python.RevertableDict:
        if type(s) is not renpy.python.StoreDeleted:
            nopickles.warnings.append((5,'Discarded students_in_play',s))
        return renpy.python.deleted
    for k,v in s.iteritems():
        ret[sanitize.expect_displayable_text(k)] = expect_Student_instance(v)
    return ret

def scrub_supply_closet(s):
    if type(s) is renpy.store.SupplyCloset:
        return s
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    nopickles.warnings.append((5,'Discarded supply_closet',s))
    return None

def scrub_training_room(s):
    if type(s) is renpy.store.TrainingRoom:
        return s
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    nopickles.warnings.append((5,'Discarded training_room',s))
    return None

def scrub_underground_case(s):
    if isinstance(s,renpy.store.DownInTheUnderground):
        return s
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    nopickles.warnings.append((5,'Discarded underground_case',s))
    return None

def scrub_vonne_path(s):
    if s is None:
        return None
    if s=='romance' or s=='friends':
        return str(s)
    if s is not False:
        nopickles.warnings.append((5,'scrub_vonne_path discarded',s))
    return False

def scrub_persistent_steam_stats(s):
    if type(s) is not dict:
        nopickles.warnings.append((5,"Discarded unexpected Steam stats.",s))
        return None
    ret = dict()
    for userid, stats in s.iteritems():
        if type(userid) is not int and type(userid) is not long:
            nopickles.warnings.append((5,'Discarded unexpected Steam userid',userid))
            continue
        if type(stats) is not dict:
            nopickles.warnings.append((5,'Discarded unexpected Steam stats',stats))
            continue
        ret[userid] = dict()
        for name, val in stats.iteritems():
            if type(name) is str and type(val) is int:
                ret[userid][name] = val
            else:
                nopickles.warnings.append((5,'Discarded unexpected Steam stat',name,val))
    return ret

